/**
 * GSweeper controller
 *
 * @author Manuke
 */

package gsweeper;

// import

import java.awt.event.*;
import java.net.*;

import javax.swing.*;

// class implementation

class GSController {
	// constant

	public static final int TIMES_MAX = 999;
	public static final int TIMER_ELAPSE = 100;

	// attribute

	GSGame game_;
	GSPictureManager pictman_;
	GSToolPanel tools_;
	GSLogoPanel logos_;
	GSInfoPanel infos_;
	GSCellPanel cells_;
	GSPictureView pictureView_;

	GSURIProcessor uriProcessor_;
	URI nextURI_;

	Long randomSeed_;

	int curTimes_ = -1;
	long firstMSec_;
	Timer timer_;

	// create & destroy

	public GSController(
		GSGame game,
		GSPictureManager pictman,
		GSToolPanel tools,
		GSLogoPanel logos,
		GSInfoPanel infos,
		GSCellPanel cells,
		GSPictureView pictureView,
		GSURIProcessor uriProcessor, URI nextURI,
		Long randomSeed)
	{
		game_ = game;
		pictman_ = pictman;
		tools_ = tools;
		logos_ = logos;
		infos_ = infos;
		cells_ = cells;
		pictureView_ = pictureView;
		uriProcessor_ = uriProcessor;
		nextURI_ = nextURI;
		randomSeed_ = randomSeed;
		addStatusListener(game_);
		addStartButtonListener(tools_.getStartButton());
		addScopeButtonListener(tools_.getScopeButton());
		updateTalismansView();
		updateTimeView();
		addInfosListener(infos_);
		if ((uriProcessor_ != null) && (nextURI_ != null)) {
			GSButton nextPageButton = infos_.getNextPageButton();
			nextPageButton.setVisible(true);
			addNextPageButtonListener(nextPageButton);
		}
		addCellListener(cells_);
		addPictureViewListener(pictureView_);
		infos_.setEnabled(false);
		timer_ = new Timer(TIMER_ELAPSE, null);
		addTimerListener(timer_);
		addLoadingLisntener(pictman);
	}

	protected void addStatusListener(GSGame game) {
		game.addStatusListener(new GSGame.StatusListener() {
			// status listener

			@Override
			public void modeChanged(GSGame.StatusEvent e) {
				pictman_.invalidateCompositePicture();
				updateAllCells();
				infos_.updateVisible(e.value);
				if (e.value == GSGame.GAMEMODE_PLAYING) {
					logos_.setVisible(false);
					if (!timer_.isRunning()) {
						timer_.start();
					}
				} else {
					if (timer_.isRunning()) {
						timer_.stop();
					}
					if (e.value == GSGame.GAMEMODE_START) {
						curTimes_ = -1;
						updateTimeView();
					}
				}
			}

			@Override
			public void opensChanged(GSGame.StatusEvent e) {
				if (pictman_.getMixMode() == GSPictureManager.MIXMODE_FADE) {
					updateAllCells();
				}
			}

			@Override
			public void talismansChanged(GSGame.StatusEvent e) {
				updateTalismansView();
			}
		});
	}

	protected void addStartButtonListener(GSButton startButton) {
		startButton.addButtonListener(new GSButton.ButtonListener() {
			// button listener

			@Override
			public void buttonClicked(GSButton.ButtonEvent e) {
				game_.clear();
				showPicture(false);
			}
		});
	}

	protected void addScopeButtonListener(GSButton scopeButton) {
		scopeButton.addButtonListener(new GSButton.ButtonListener() {
			// button listener

			@Override
			public void buttonClicked(GSButton.ButtonEvent e) {
				showPicture(e.stat);
			}
		});
	}

	protected void addInfosListener(GSInfoPanel infos) {
		infos.addPanelListener(new GSClickablePanel.PanelListener() {
			// panel listener

			@Override
			public void panelClicked(GSClickablePanel.PanelEvent e) {
				showPicture(true);
			}
		});
	}

	protected void addNextPageButtonListener(GSButton nextPageButton) {
		nextPageButton.addButtonListener(new GSButton.ButtonListener() {
			// button listener

			@Override
			public void buttonClicked(GSButton.ButtonEvent e) {
				uriProcessor_.browseURI(nextURI_);
			}
		});
	}

	protected void addCellListener(GSCellPanel cellButton) {
		cellButton.addCellListener(new GSCellPanel.CellListener() {
			// cell listener

			@Override
			public void cellOpened(GSCellPanel.CellEvent e) {
				if (game_.getGameMode() <= GSGame.GAMEMODE_START) {
					if (randomSeed_ == null) {
						game_.start(e.col, e.row);
					} else {
						game_.start(e.col, e.row, randomSeed_);
					}
				}
				game_.openCell(e.col, e.row);
			}

			@Override
			public void cellMarked(GSCellPanel.CellEvent e) {
				if (game_.getGameMode() <= GSGame.GAMEMODE_START) {
					return;
				}
				game_.markCell(e.col, e.row);
			}

			@Override
			public void cellExOpened(GSCellPanel.CellEvent e) {
				game_.openCellEx(e.col, e.row);
			}

			@Override
			public void cellExOpening(GSCellPanel.CellEvent e) {
				game_.prepareCell(e.col, e.row, e.stat);
			}
		});
	}

	protected void addPictureViewListener(GSPictureView pictureView) {
		pictureView.addPanelListener(new GSClickablePanel.PanelListener() {
			// panel listener

			@Override
			public void panelClicked(GSClickablePanel.PanelEvent e) {
				showPicture(false);
			}
		});
	}

	protected void addTimerListener(Timer timer) {
		timer_.addActionListener(new ActionListener() {
			// timer listener

			@Override
			public void actionPerformed(ActionEvent e) {
				if (game_.getGameMode() == GSGame.GAMEMODE_PLAYING) {
					long nowMSec = e.getWhen();
					if (curTimes_ < 0) {
						firstMSec_ = nowMSec;
					}
					curTimes_ = (int)((nowMSec-firstMSec_)/1000);
					updateTimeView();
					if ((game_.getTimes() >= 0) &&
						(curTimes_ >= game_.getTimes()))
					{
						game_.timeLimit();
						showPicture(false);
					}
				}
			}
		});
	}

	protected void addLoadingLisntener(GSPictureManager pictman) {
		pictman.addLoadingListener(new GSPictureManager.LoadingListener() {
			// loading listener

			@Override
			public void loadingUpdated(GSPictureManager.LoadingEvent e) {
				if (e.completed) {
					infos_.updateVisible(game_.getGameMode());
					infos_.setEnabled(true);
					updateAllCells();
				}
			}
		});
	}

	// operation

	public void showPicture(boolean visible) {
		cells_.setVisible(!visible);
		tools_.getScopeButton().setPushed(visible);
		logos_.setVisible(false);
		infos_.updateVisible(visible? -1: game_.getGameMode());
		pictureView_.setVisible(visible);
	}

	public void updateTalismansView() {
		tools_.getTalismansView().setNumber(
			game_.getGhosts()-game_.getTalismans());
	}

	public void updateTimeView() {
		int times = game_.getTimes();
		if (times < 0) {
			times = (curTimes_ >= 0)? curTimes_: 0;
			if (times > TIMES_MAX) {
				times = TIMES_MAX;
			}
		} else {
			times -= (curTimes_ >= 0)? curTimes_: 0;
			if (times < 0) {
				times = 0;
			}
		}
		tools_.getTimeView().setNumber(times);
	}

	public void updateAllCells() {
		if (cells_.isVisible()) {
			cells_.setAllDirty();
			cells_.repaint();
		}
	}
}
