/**
 * GSweeper game model
 *
 * @author Manuke
 */

package gsweeper;

// import

import java.awt.*;
import java.util.*;

import javax.swing.event.*;

// class implementation

class GSGame {
	// type

	public class StatusEvent
		extends EventObject
	{
		// attribute

		public int value;

		// create & destroy

		public StatusEvent(
			Object source,
			int value)
		{
			super(source);
			this.value = value;
		}
	}

	public interface StatusListener
		extends EventListener
	{
		// status listener

		public void modeChanged(StatusEvent e);
		public void opensChanged(StatusEvent e);
		public void talismansChanged(StatusEvent e);
	}

	public class MapEvent
		extends EventObject
	{
		// attribute

		public int col;
		public int row;

		// create & destroy

		public MapEvent(
			Object source,
			int col, int row)
		{
			super(source);
			this.col = col;
			this.row = row;
		}
	}

	public interface MapListener
		extends EventListener
	{
		// map listener

		public int getCol();
		public int getRow();
		public void mapChanged(MapEvent e);
	}

	// constant

	public static final int LEVEL_EASY = 0;
	public static final int LEVEL_NORMAL = 1;
	public static final int LEVEL_HARD = 2;

	public static final int RECOMMEND_TOTALS_EASY = 81;
	public static final int RECOMMEND_GHOSTS_EASY = 10;
	public static final int RECOMMEND_TOTALS_NORMAL = 256;
	public static final int RECOMMEND_GHOSTS_NORMAL = 40;
	public static final int RECOMMEND_TOTALS_HARD = 480;
	public static final int RECOMMEND_GHOSTS_HARD = 99;

	public static final int GAMEMODE_START = 0;
	public static final int GAMEMODE_PLAYING = 1;
	public static final int GAMEMODE_COMPLETED = 2;
	public static final int GAMEMODE_CURSED = 3;
	public static final int GAMEMODE_TIMELIMIT = 4;

	public static final int AROUND_MASK = 0x000F;
	public static final int GHOST       = 0x0010;
	public static final int TALISMAN    = 0x0020;
	public static final int QUESTION    = 0x0040;
	public static final int PREPARED    = 0x0080;
	public static final int OPENED      = 0x0100;

	// attribute

	int level_;
	int cols_;
	int rows_;
	int ghosts_;
	int times_;
	int gameMode_ = GAMEMODE_START;
	int opens_ = 0;
	int talismans_ = 0;
	int[][] map_;
	Dimension pictureDim_ = new Dimension(1, 1);

	EventListenerList mapListeners_ =
		new EventListenerList();
	EventListenerList statusListeners_ =
		new EventListenerList();

	// create & destroy

	public GSGame(
		int level,
		int cols, int rows, int ghosts,
		int times,
		Dimension pictureDim)
	{
		level_ = level;
		cols_ = cols;
		rows_ = rows;
		ghosts_ = ghosts;
		times_ = times;
		pictureDim_.width = pictureDim.width;
		pictureDim_.height = pictureDim.height;
		double ratio = 1.0;
		if ((pictureDim_.width > 0) &&
			(pictureDim_.height > 0))
		{
			ratio =
				(double)pictureDim_.width/
					pictureDim_.height;
		}
		int recommendTotals, recommendGhosts;
		if (level_ == LEVEL_EASY) {
			recommendTotals = RECOMMEND_TOTALS_EASY;
			recommendGhosts = RECOMMEND_GHOSTS_EASY;
		} else if (level_ == LEVEL_NORMAL) {
			recommendTotals = RECOMMEND_TOTALS_NORMAL;
			recommendGhosts = RECOMMEND_GHOSTS_NORMAL;
		} else {
			recommendTotals = RECOMMEND_TOTALS_HARD;
			recommendGhosts = RECOMMEND_GHOSTS_HARD;
		}
		if ((cols_ <= 0) && (rows_ <= 0)) {
			cols_ = (int)(Math.sqrt(recommendTotals*ratio)+0.5);
			rows_ = (int)(Math.sqrt(recommendTotals/ratio)+0.5);
		} else if (cols_ <= 0) {
			cols_ = (int)(rows_*ratio+0.5);
		} else if (rows_ <= 0) {
			rows_ = (int)(cols_/ratio+0.5);
		}
		if (cols_ < 1) {
			cols_ = 1;
		}
		if (rows_ < 1) {
			rows_ = 1;
		}
		int totals = cols_*rows_;
		if (ghosts_ < 0) {
			ghosts_ = (int)(
				(double)totals*recommendGhosts/
					recommendTotals+0.5);
			if (ghosts_ < 0) {
				ghosts_ = 0;
			}
		} else if (ghosts_ >= totals) {
			ghosts_ = totals-1;
		}
		map_ = new int[rows_][];
		for (int row = 0; row < rows_; row++) {
			map_[row] = new int[cols_];
		}
		clear();
	}

	public void clear() {
		for (int row = 0; row < rows_; row++) {
			for (int col = 0; col < cols_; col++) {
				setMapValue(col, row, 0);
			}
		}
		changeGameMode(GAMEMODE_START);
		if (opens_ != 0) {
			opens_ = 0;
			fireOpensChanged();
		}
		if (talismans_ != 0) {
			talismans_ = 0;
			fireTalismansChanged();
		}
	}

	// property

	public int getLevel() {
		return level_;
	}

	public int getCols() {
		return cols_;
	}

	public int getRows() {
		return rows_;
	}

	public int getTotals() {
		return cols_*rows_;
	}

	public int getGhosts() {
		return ghosts_;
	}

	public int getTimes() {
		return times_;
	}

	int getGameMode() {
		return gameMode_;
	}

	public int getOpens() {
		return opens_;
	}

	public int getBlank() {
		return getTotals()-getGhosts();
	}

	public int getTalismans() {
		return talismans_;
	}

	public Dimension getPictureDim() {
		return new Dimension(pictureDim_);
	}

	// map

	public boolean isInside(int col, int row) {
		return (col >= 0) && (col < cols_) &&
			(row >= 0) && (row < rows_);
	}
	public int getMapValue(int col, int row) {
		int value = 0;
		if (isInside(col, row)) {
			value = map_[row][col];
		}
		return value;
	}
	public boolean setMapValue(int col, int row, int value) {
		boolean changed = false;
		if (isInside(col, row) &&
			(value != map_[row][col]))
		{
			changed = true;
			map_[row][col] = value;
			fireMapChanged(col, row);
		}
		return changed;
	}
	public int getAroundGhosts(int col, int row) {
		return getMapValue(col, row) & AROUND_MASK;
	}
	public boolean isMapFlag(int col, int row, int flag) {
		return (getMapValue(col, row) & flag) != 0;
	}
	public boolean setMapFlag(int col, int row, int flag, boolean stat) {
		int value = getMapValue(col, row);
		if (stat) {
			value |= flag;
		} else {
			value &= ~flag;
		}
		return setMapValue(col, row, value);
	}
	public boolean isGhost(int col, int row) {
		return isMapFlag(col, row, GHOST);
	}
	public boolean isTalisman(int col, int row) {
		return isMapFlag(col, row, TALISMAN);
	}
	public boolean isQuestion(int col, int row) {
		return isMapFlag(col, row, QUESTION);
	}
	public boolean isPrepared(int col, int row) {
		return isMapFlag(col, row, PREPARED);
	}
	public boolean isOpened(int col, int row) {
		return isMapFlag(col, row, OPENED);
	}
	public int getAroundTalismans(int col, int row) {
		int aroundTalismans = 0;
		for (int row2 = row-1; row2 <= row+1; row2++) {
			for (int col2 = col-1; col2 <= col+1; col2++) {
				if (isTalisman(col2, row2)) {
					aroundTalismans++;
				}
			}
		}
		return aroundTalismans;
	}

	// operation

	public void changeGameMode(int gameMode) {
		if (gameMode != gameMode_) {
			gameMode_ = gameMode;
			fireModeChanged();
		}
	}

	public void start(int exCol, int exRow, long seed) {
		clear();
		int totals = getTotals(),
			shuffles = totals;
		if (isInside(exCol, exRow)) {
			shuffles--;
		}
		int[][] seq = new int[totals][];
		Random rnd = new Random(seed);
		for (int row = 0, i = 0; row < rows_; row++) {
			for (int col = 0; col < cols_; col++) {
				if ((col != exCol) || (row != exRow)) {
					seq[i++] = new int[] { col, row };
				}
			}
		}
		if (shuffles < totals) {
			seq[shuffles] = new int[] { exCol, exRow };
		}
		for (int i = 0; i < shuffles; i++) {
			int j = rnd.nextInt(shuffles);
			int[] tmp = seq[i];
			seq[i] = seq[j];
			seq[j] = tmp;
		}
		int n;
		for (n = 0; n < ghosts_; n++) {
			setMapFlag(seq[n][0], seq[n][1], GHOST, true);
		}
		for (; n < totals; n++) {
			int col = seq[n][0], row = seq[n][1];
			int aroundGhosts = 0;
			for (int row2 = row-1; row2 <= row+1; row2++) {
				for (int col2 = col-1; col2 <= col+1; col2++) {
					if (isGhost(col2, row2)) {
						aroundGhosts++;
					}
				}
			}
			setMapValue(col, row, aroundGhosts);
		}
		changeGameMode(GAMEMODE_PLAYING);
	}

	public void start(int exCol, int exRow) {
		start(exCol, exRow, System.currentTimeMillis());
	}

	public boolean openCell(int col, int row) {
		if (!isInside(col, row) ||
			isOpened(col, row) ||
			isTalisman(col, row))
		{
			return false;
		}
		if (isGhost(col, row)) {
			curse();
			return true;
		}
		setMapFlag(col, row, OPENED, true);
		opens_++;
		fireOpensChanged();
		if (getOpens() == getBlank()) {
			complete();
			return true;
		}
		if (getAroundGhosts(col, row) != 0) {
			return false;
		}
		LinkedList<int[]> openQueue = new LinkedList<int[]>();
		openQueue.add(new int[] { col, row });
		while (openQueue.size() > 0) {
			int[] cell = openQueue.pop();
			for (int row2 = cell[1]-1; row2 <= cell[1]+1; row2++) {
				for (int col2 = cell[0]-1; col2 <= cell[0]+1; col2++) {
					if (((col2 == cell[0]) && (row2 == cell[1])) ||
						!isInside(col2, row2) ||
						isOpened(col2, row2))
					{
						continue;
					}
					setMapFlag(col2, row2, OPENED, true);
					opens_++;
					fireOpensChanged();
					if (getOpens() == getBlank()) {
						complete();
						return true;
					}
					if (getAroundGhosts(col2, row2) == 0) {
						openQueue.add(new int[] { col2, row2 });
					}
				}
			}
		}
		return false;
	}

	public boolean openCellEx(int col, int row) {
		if (!isOpened(col, row) ||
			(getAroundGhosts(col, row) !=
				getAroundTalismans(col, row)))
		{
			return false;
		}
		for (int row2 = row-1; row2 <= row+1; row2++) {
			for (int col2 = col-1; col2 <= col+1; col2++) {
				if (((col2 != col) || (row2 != row)) &&
					isInside(col2, row2) &&
					!isOpened(col2, row2) &&
					(isGhost(col2, row2) !=
						isTalisman(col2, row2)))
				{
					curse();
					return true;
				}
			}
		}
		for (int row2 = row-1; row2 <= row+1; row2++) {
			for (int col2 = col-1; col2 <= col+1; col2++) {
				if (((col2 != col) || (row2 != row)) &&
					isInside(col2, row2) &&
					!isOpened(col2, row2) &&
					!isTalisman(col2, row2))
				{
					if (openCell(col2, row2)) {
						return true;
					}
				}
			}
		}
		return false;
	}

	public void markTalisman(int col, int row, boolean stat) {
		if (!isInside(col, row)) {
			return;
		}
		setMapFlag(col, row, QUESTION, false);
		if (setMapFlag(col, row, TALISMAN, stat)) {
			talismans_ += stat? 1: -1;
			fireTalismansChanged();
		}
	}

	public void markQuestion(int col, int row, boolean stat) {
		if (!isInside(col, row)) {
			return;
		}
		if (setMapFlag(col, row, TALISMAN, false)) {
			talismans_--;
			fireTalismansChanged();
		}
		setMapFlag(col, row, QUESTION, stat);
	}

	public void markCell(int col, int row) {
		if (!isInside(col, row) ||
			isOpened(col, row))
		{
			return;
		}
		if (isQuestion(col, row)) {
			markQuestion(col, row, false);
		} else if (
			isTalisman(col, row) ||
			(getTalismans() >= getGhosts()))
		{
			markTalisman(col, row, false);
			markQuestion(col, row, true);
		} else {
			markQuestion(col, row, false);
			markTalisman(col, row, true);
		}
	}

	public void prepareCell(int col, int row, boolean stat) {
		for (int row2 = row-1; row2 <= row+1; row2++) {
			for (int col2 = col-1; col2 <= col+1; col2++) {
				if (((col2 != col) || (row2 != row)) &&
					isInside(col2, row2) &&
					!isOpened(col2, row2) &&
					!isTalisman(col2, row2))
				{
					setMapFlag(col2, row2, PREPARED, stat);
				}
			}
		}
	}

	public void complete() {
		for (int row = 0; row < getRows(); row++) {
			for (int col = 0; col < getCols(); col++) {
				if (!isOpened(col, row) &&
					isGhost(col, row))
				{
					markTalisman(col, row, true);
					setMapFlag(col, row, OPENED, true);
				}
			}
		}
		changeGameMode(GAMEMODE_COMPLETED);
	}

	public void curse() {
		for (int row = 0; row < getRows(); row++) {
			for (int col = 0; col < getCols(); col++) {
				if (!isOpened(col, row) &&
					(isGhost(col, row) != isTalisman(col, row)))
				{
					setMapFlag(col, row, OPENED, true);
				}
			}
		}
		changeGameMode(GAMEMODE_CURSED);
	}

	public void timeLimit() {
		for (int row = 0; row < getRows(); row++) {
			for (int col = 0; col < getCols(); col++) {
				if (!isOpened(col, row) &&
					(isGhost(col, row) != isTalisman(col, row)))
				{
					setMapFlag(col, row, OPENED, true);
				}
			}
		}
		changeGameMode(GAMEMODE_TIMELIMIT);
	}

	// utility

	public Rectangle getCellRect(int col, int row) {
		int x1 = (pictureDim_.width*col)/cols_,
			y1 = (pictureDim_.height*row)/rows_,
			x2 = (pictureDim_.width*(col+1))/cols_,
			y2 = (pictureDim_.height*(row+1))/rows_;
		return new Rectangle(x1, y1, x2-x1, y2-y1);
	}

	// status event

	public void addStatusListener(StatusListener listener) {
		statusListeners_.add(StatusListener.class, listener);
	}

	public void removeStatusListener(StatusListener listener) {
		statusListeners_.remove(StatusListener.class, listener);
	}

	void fireModeChanged() {
		StatusEvent e = new StatusEvent(this, gameMode_);
		for (
			StatusListener listener :
				statusListeners_.getListeners(StatusListener.class))
		{
			listener.modeChanged(e);
		}
	}

	void fireOpensChanged() {
		StatusEvent e = new StatusEvent(this, opens_);
		for (
			StatusListener listener :
				statusListeners_.getListeners(StatusListener.class))
		{
			listener.opensChanged(e);
		}
	}

	void fireTalismansChanged() {
		StatusEvent e = new StatusEvent(this, talismans_);
		for (
			StatusListener listener :
				statusListeners_.getListeners(StatusListener.class))
		{
			listener.talismansChanged(e);
		}
	}

	// map event

	public void addMapListener(MapListener listener) {
		mapListeners_.add(MapListener.class, listener);
	}

	public void removeMapListener(MapListener listener) {
		mapListeners_.remove(MapListener.class, listener);
	}

	void fireMapChanged(int col, int row) {
		MapEvent e = new MapEvent(this, col, row);
		for (
			MapListener listener :
				mapListeners_.getListeners(MapListener.class))
		{
			int col2 = listener.getCol(),
				row2 = listener.getRow();
			if (((col2 < 0) || (col2 == col)) &&
				((row2 < 0) || (row2 == row)))
			{
				listener.mapChanged(e);
			}
		}
	}
}
