/**
 * GSweeper parameters parser
 *
 * @author Manuke
 */

package gsweeper;

// import

import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;

import getopt.*;

// class implementation

class GSParameters {
	// constant

	private static final String[][] PARAMETER_DATA = {
		{
			"h", "help",
			null, null,
			null, "show help(this message)"
		}, {
			"l", "level",
			"string", "<level>",
			"game level(easy/normal/hard, default: normal)", null
		}, {
			"c", "cols",
			"integer", "N",
			"cell column count(default: auto)", null
		}, {
			"r", "rows",
			"integer", "N",
			"cell row count(default: auto)", null
		}, {
			"g", "ghosts",
			"integer", "N",
			"ghost count(default: auto)", null
		}, {
			"t", "times",
			"integer", "N",
			"time limit in seconds(default: no limit)", null
		}, {
			"m", "mixmode",
			"string", "<mode>",
			"picture mixing mode(block/fade, default: block)", null
		}, {
			"1", "picture1",
			"url", "<filename>",
			"picture1 url", "picture1 filename"
		}, {
			"2", "picture2",
			"url", "<filename>",
			"picture2 url", "picture2 filename"
		}, {
			"n", "next",
			"url", "<url>",
			"next page", null
		}, {
			"s", "seed",
			"integer", "N",
			"random seed(default: used current time)", null
		}
	};

	public static final String DEFAULT_RESOURCE_NAME = "text/params.txt";

	public static final String APP_NAME = "GhostSweeper";
	public static final String VERSION = "1.00";
	public static final String AUTHOR = "Manuke";

	public static final String APP_INFO;
	public static final String SHORT_OPTS;
	public static final String[] LONG_OPTS;
	public static final String[][] PARAMETER_INFO;
	public static final String[] HELP_MESSAGE;

	// attribute

	public boolean help = false;
	public int level = GSGame.LEVEL_NORMAL;
	public int cols = -1;
	public int rows = -1;
	public int ghosts = -1;
	public int times = -1;
	public int mixMode = GSPictureManager.MIXMODE_BLOCK;
	public String picture1 = null;
	public URI picture1URI = null;
	public String picture2 = null;
	public URI picture2URI = null;
	public String next = null;
	public URI nextURI = null;
	public Long randomSeed = null;

	public GSURIProcessor uriProcessor = null;

	// create & destroy

	static {
		APP_INFO = String.format(
			"%s Ver. %s, Programmed by %s",
			APP_NAME, VERSION, AUTHOR);
		ArrayList<String[]> parameterInfo =
			new ArrayList<String[]>();
		StringBuffer shortOpts =
			new StringBuffer();
		ArrayList<String> longOpts =
			new ArrayList<String>();
		ArrayList<String> helpMessage =
			new ArrayList<String>();
		helpMessage.add(APP_INFO);
		helpMessage.add("");
		helpMessage.add("usage: java -jar gsweeper.jar [opt...]");
		helpMessage.add("");
		helpMessage.add("Options:");
		for (String[] optInfo : PARAMETER_DATA) {
			String sopt = optInfo[0],
				lopt = optInfo[1],
				paramType = optInfo[2],
				paramName = optInfo[3],
				descApplet = optInfo[4],
				descCommand = optInfo[5];
			if ((descApplet == null) && (descCommand != null)) {
				descApplet = descCommand;
			} else if ((descApplet != null) && (descCommand == null)) {
				descCommand = descApplet;
			}
			boolean hasParam = (paramName != null);
			String helpLine = "    ";
			if (sopt != null) {
				String sopt2 = sopt;
				helpLine += "-"+sopt;
				if (hasParam) {
					sopt2 += ":";
					helpLine += " "+paramName;
				}
				shortOpts.append(sopt2);
			}
			if (lopt != null) {
				String lopt2 = lopt;
				if (sopt != null) {
					helpLine += ", ";
				}
				helpLine += "--"+lopt;
				if (hasParam) {
					lopt2 += "=";
					helpLine += "="+paramName;
				}
				longOpts.add(lopt2);
				if (paramType != null) {
					parameterInfo.add(new String[] {
						lopt, paramType, descApplet
					});
				}
			}
			helpMessage.add(helpLine);
			helpMessage.add("        "+descCommand);
		}
		SHORT_OPTS = shortOpts.toString();
		LONG_OPTS = longOpts.toArray(new String[0]);
		PARAMETER_INFO = parameterInfo.toArray(new String[0][0]);
		HELP_MESSAGE = helpMessage.toArray(new String[0]);
	}

	// decode

	public void decode(
		Iterable<String[]> opts,
		Iterable<String> args)
	{
		for (String[] opt : opts) {
			if (
				opt[0].equals("-h") ||
				opt[0].equals("--help"))
			{
				help = true;
			} else if (
				opt[0].equals("-l") ||
				opt[0].equals("--level"))
			{
				if (opt[1].equalsIgnoreCase("e") ||
					opt[1].equalsIgnoreCase("easy") ||
					opt[1].equalsIgnoreCase("1"))
				{
					level = GSGame.LEVEL_EASY;
				} else if (
					opt[1].equalsIgnoreCase("n") ||
					opt[1].equalsIgnoreCase("normal") ||
					opt[1].equalsIgnoreCase("2"))
				{
					level = GSGame.LEVEL_NORMAL;
				} else if (
					opt[1].equalsIgnoreCase("h") ||
					opt[1].equalsIgnoreCase("hard") ||
					opt[1].equalsIgnoreCase("3"))
				{
					level = GSGame.LEVEL_HARD;
				}
			} else if (
				opt[0].equals("-c") ||
				opt[0].equals("--cols"))
			{
				try {
					cols = Integer.valueOf(opt[1]);
				} catch (NumberFormatException e) {
				}
			} else if (
				opt[0].equals("-r") ||
				opt[0].equals("--rows"))
			{
				try {
					rows = Integer.valueOf(opt[1]);
				} catch (NumberFormatException e) {
				}
			} else if (
				opt[0].equals("-g") ||
				opt[0].equals("--ghosts"))
			{
				try {
					ghosts = Integer.valueOf(opt[1]);
				} catch (NumberFormatException e) {
				}
			} else if (
				opt[0].equals("-t") ||
				opt[0].equals("--times"))
			{
				try {
					times = Integer.valueOf(opt[1]);
				} catch (NumberFormatException e) {
				}
			} else if (
				opt[0].equals("-m") ||
				opt[0].equals("--mixmode"))
			{
				if (
					opt[1].equalsIgnoreCase("b") ||
					opt[1].equalsIgnoreCase("block"))
				{
					mixMode = GSPictureManager.MIXMODE_BLOCK;
				} else if (
					opt[1].equalsIgnoreCase("f") ||
					opt[1].equalsIgnoreCase("fade"))
				{
					mixMode = GSPictureManager.MIXMODE_FADE;
				}
			} else if (
				opt[0].equals("-1") ||
				opt[0].equals("--picture1"))
			{
				picture1 = opt[1];
				if (uriProcessor != null) {
					picture1URI =
						uriProcessor.createURI(picture1);
				}
			} else if (
				opt[0].equals("-2") ||
				opt[0].equals("--picture2"))
			{
				picture2 = opt[1];
				if (uriProcessor != null) {
					picture2URI =
						uriProcessor.createURI(picture2);
				}
			} else if (
				opt[0].equals("-n") ||
				opt[0].equals("--next"))
			{
				next = opt[1];
				if (uriProcessor != null) {
					nextURI =
						uriProcessor.createURI(next);
				}
			} else if (
				opt[0].equals("-s") ||
				opt[0].equals("--seed"))
			{
				try {
					randomSeed = Long.valueOf(opt[1]);
				} catch (NumberFormatException e) {
				}
			}
		}
	}

	public void decode(String[] args) {
		GetOpt getopt = new GetOpt(SHORT_OPTS, LONG_OPTS);
		GetOpt.Result goresult = getopt.decode(args);
		decode(goresult.opts, goresult.args);
	}

	// resource

	public boolean loadResource(String name) {
		LinkedList<String> args =
			new LinkedList<String>();
		try {
			URL url = GSParameters.class.
				getClassLoader().getResource(name);
			if (url != null) {
				BufferedReader reader = new BufferedReader(
					new InputStreamReader(
						url.openStream(),
						"UTF-8"));
				for (
					String line = reader.readLine();
					line != null;
					line = reader.readLine())
				{
					args.add(line.replaceAll("[\r\n]", ""));
				}
				reader.close();
			}
		} catch (IOException e) {
		}
		if (args.size() <= 0) {
			return false;
		}
		GSURIProcessor orgURIProcessor = uriProcessor;
		uriProcessor = new GSURIProcessor() {
			// uri

			public URI createURI(String uriStr) {
				if ((uriStr == null) || (uriStr.length() <= 0)) {
					return null;
				}
				URI uri;
				try {
					URL url = GSParameters.class.
							getClassLoader().getResource(uriStr);
					uri = url.toURI();
				} catch (Exception e) {
					uri = null;
				}
				return uri;
			}

			public void browseURI(URI uri) {
				try {
					Desktop.getDesktop().browse(uri);
				} catch (Exception e) {
				}
			}
		};
		decode(args.toArray(new String[0]));
		uriProcessor = orgURIProcessor;
		return true;
	}

	public boolean loadResource() {
		return loadResource(DEFAULT_RESOURCE_NAME);
	}
}
