/**
 * GSweeper picture image manager
 *
 * @author Manuke
 */

package gsweeper;

// import

import java.awt.*;
import java.awt.image.*;
import java.awt.color.*;
import java.net.*;
import java.util.*;

import javax.imageio.*;
import javax.imageio.event.*;
import javax.imageio.stream.*;
import javax.swing.event.*;

// class implementation

class GSPictureManager {
	// type

	public class LoadingEvent
		extends EventObject
	{
		// attribute

		float step;
		boolean completed;

		// create & destroy

		public LoadingEvent(
			Object source,
			float step, boolean completed)
		{
			super(source);
			this.step = step;
			this.completed = completed;
		}
	}

	public interface LoadingListener
		extends EventListener
	{
		// loading listener

		public void loadingUpdated(LoadingEvent e);
	}

	// constant

	public static final int MIXMODE_BLOCK = 0;
	public static final int MIXMODE_FADE = 1;

	public static final int CX_DEFAULT_PICTURE = 400;
	public static final int CY_DEFAULT_PICTURE = 400;

	public static Color DEFAULT_PICTURE_COLOR1 =
		new Color(0x40FFFF);
	public static Color DEFAULT_PICTURE_COLOR2 =
		new Color(0x40FF80);

	// attribute

	Object sync_ = new Object();
	Object sync2_ = new Object();
	URI[] pictureURIs_ = new URI[2];
	Dimension pictureDim_ = new Dimension(
		CX_DEFAULT_PICTURE, CY_DEFAULT_PICTURE);
	BufferedImage[] pictures_ = {
		null, null
	};

	int loadingIndex_ = 0;
	ImageReader reader_ = null;
	float progressBase_ = 0f;
	float progressStep_ = 0.5f;

	EventListenerList loadingListeners_ =
		new EventListenerList();
	boolean completed_ = false;

	int mixMode_;
	BufferedImage compositePicture_ = null;
	int opens_ = -1;

	// create & destroy

	public GSPictureManager(
		URI picture1URI, URI picture2URI,
		int mixMode)
	{
		pictureURIs_[0] = picture1URI;
		pictureURIs_[1] = picture2URI;
		mixMode_ = mixMode;
		loadingIndex_ = 0;
		reader_ = setupReader(
			pictureURIs_[loadingIndex_], pictureDim_);
		if (reader_ == null) {
			loadingIndex_++;
			reader_ = setupReader(
				pictureURIs_[loadingIndex_], pictureDim_);
		}
		compositePicture_ = new BufferedImage(
			pictureDim_.width,
			pictureDim_.height,
			BufferedImage.TYPE_INT_ARGB);
		Thread thread = new Thread() {
			// thread

			@Override
			public void run() {
				loadAllPictures();
			}
		};
		thread.start();
	}

	// property

	public Dimension getPictureDim() {
		return new Dimension(pictureDim_);
	}

	public BufferedImage getPicture(int index) {
		BufferedImage picture = null;
		if ((index >= 0) && (index < 2)) {
			synchronized (sync_) {
				picture = pictures_[index];
			}
		}
		return picture;
	}

	int getMixMode() {
		return mixMode_;
	}

	// load

	protected ImageReader setupReader(
		URI pictureURI,
		Dimension pictureDim)
	{
		if (pictureURI == null) {
			return null;
		}
		ImageReader reader = null;
		try {
			ImageInputStream is =
				ImageIO.createImageInputStream(
					pictureURI.toURL().openStream());
			for (
				Iterator<ImageReader> it =
					ImageIO.getImageReaders(is);
				it.hasNext();
				)
			{
				reader = it.next();
				try {
					reader.setInput(is);
					int width = reader.getWidth(0),
						height = reader.getHeight(0);
					if (pictureDim != null) {
						pictureDim.width = width;
						pictureDim.height = height;
					}
					break;
				} catch (Exception e) {
					if (reader != null) {
						reader.dispose();
						reader = null;
					}
				}
			}
		} catch (Exception e) {
			if (reader != null) {
				reader.dispose();
				reader = null;
			}
		}
		return reader;
	}

	protected void loadAllPictures() {
		fireLoadingUpdated(0f, false);
		progressBase_ = 0f;
		progressStep_ = 0.5f;
		if ((pictureURIs_[0] == null) ||
			(pictureURIs_[1] == null))
		{
			progressStep_ = 1f;
		}
		for (; loadingIndex_ < 2; loadingIndex_++) {
			if (reader_ == null) {
				reader_ = setupReader(
					pictureURIs_[loadingIndex_], null);
			}
			if (reader_ != null) {
				addReadProgressListener(reader_);
				BufferedImage picture;
				try {
					picture = reader_.read(0);
				} catch (Exception e) {
					picture = null;
				}
				synchronized (sync_) {
					pictures_[loadingIndex_] = picture;
				}
				reader_.dispose();
				reader_ = null;
			}
			progressBase_ += progressStep_;
		}
		buildDefaultPictures();
		synchronized (sync2_) {
			opens_ = -1;
		}
		fireLoadingUpdated(1f, true);
	}

	protected void addReadProgressListener(
		ImageReader reader)
	{
		reader.addIIOReadProgressListener(
			new IIOReadProgressListener()
		{
			// read progress

			@Override
			public void thumbnailStarted(
				ImageReader source,
				int imageIndex,
				int thumbnailIndex)
			{
			}

			@Override
			public void thumbnailProgress(
				ImageReader source,
				float percentageDone)
			{
			}

			@Override
			public void thumbnailComplete(
				ImageReader source)
			{
			}

			@Override
			public void sequenceStarted(
				ImageReader source,
				int minIndex)
			{
			}

			@Override
			public void sequenceComplete(
				ImageReader source)
			{
			}

			@Override
			public void readAborted(
				ImageReader source)
			{
			}

			@Override
			public void imageStarted(
				ImageReader source,
				int imageIndex)
			{
			}

			@Override
			public void imageProgress(
				ImageReader source,
				float percentageDone)
			{
				fireLoadingUpdated(
					progressBase_+
						progressStep_*
							(percentageDone/100f),
					false);
			}

			@Override
			public void imageComplete(
				ImageReader source)
			{
			}
		});
	}


	// build

	protected void buildDefaultPictures() {
		BufferedImage picture1 = getPicture(0),
			picture2 = getPicture(1);
		if ((picture1 == null) && (picture2 == null)) {
			picture2 = new BufferedImage(
				pictureDim_.width, pictureDim_.height,
				BufferedImage.TYPE_INT_ARGB);
			Graphics g = picture2.getGraphics();
			if (g instanceof Graphics2D) {
				Graphics2D g2 = (Graphics2D)g;
				g2.setPaint(
					new GradientPaint(
						0, 0, DEFAULT_PICTURE_COLOR1,
						0, pictureDim_.height, DEFAULT_PICTURE_COLOR2));
			} else {
				g.setColor(DEFAULT_PICTURE_COLOR1);
			}
			g.fillRect(0, 0, pictureDim_.width, pictureDim_.height);
			g.dispose();
		}
		if ((picture1 == null) || (picture2 == null)) {
			BufferedImage grayPicture = new BufferedImage(
				pictureDim_.width, pictureDim_.height,
				BufferedImage.TYPE_INT_ARGB);
			ColorConvertOp ccOp = new ColorConvertOp(
				ColorSpace.getInstance(ColorSpace.CS_sRGB),
				ColorSpace.getInstance(ColorSpace.CS_GRAY),
				null);
			ccOp.filter(
				(picture1 != null)? picture1: picture2,
				grayPicture);
			if (picture1 == null) {
				picture1 = grayPicture;
			} else {
				picture2 = grayPicture;
			}
		}
		synchronized (sync_) {
			pictures_[0] = picture1;
			pictures_[1] = picture2;
		}
	}

	public BufferedImage buildCompositePicture(GSGame game) {
		boolean dirty;
		synchronized (sync2_) {
			dirty = (opens_ != game.getOpens());
			opens_ = game.getOpens();
		}
		if (dirty) {
			BufferedImage picture1 = getPicture(0),
				picture2 = getPicture(1);
			int opens = game.getOpens(),
				maxOpens = game.getTotals()-game.getGhosts();
			Graphics g = compositePicture_.getGraphics();
			if ((picture1 == null) || (picture2 == null)) {
				Color orgColor = g.getColor();
				g.setColor(new Color(0x000000));
				g.fillRect(0, 0, pictureDim_.width, pictureDim_.height);
				g.setColor(orgColor);
			} else if (opens <= 0) {
				g.drawImage(picture1, 0, 0, null);
			} else if (opens >= maxOpens) {
				g.drawImage(picture2, 0, 0, null);
			} else {
				g.drawImage(picture1, 0, 0, null);
				if ((mixMode_ == MIXMODE_FADE) &&
					(g instanceof Graphics2D))
				{
					Graphics2D g2 = (Graphics2D)g;
					float alpha = 0.0f+
						((maxOpens > 0)?
							(float)opens/maxOpens:
							0.0f);
					Composite orgComposite = g2.getComposite();
					g2.setComposite(
						AlphaComposite.getInstance(
							AlphaComposite.SRC_OVER, alpha));
					g.drawImage(picture2, 0, 0, null);
					g2.setComposite(orgComposite);
				} else {
					for (int row = 0; row < game.getRows(); row++) {
						for (int col = 0; col < game.getCols(); col++) {
							if (game.isOpened(col, row) &&
								(game.isTalisman(col, row) ==
									game.isGhost(col, row)))
							{
								Rectangle cellRect = game.getCellRect(col, row);
								int x1 = cellRect.x,
									y1 = cellRect.y,
									x2 = cellRect.x+cellRect.width,
									y2 = cellRect.y+cellRect.height;
								g.drawImage(
									picture2,
									x1, y1, x2, y2,
									x1, y1, x2, y2,
									null);
							}
						}
					}
				}
			}
			g.dispose();
		}
		return compositePicture_;
	}

	public void invalidateCompositePicture() {
		synchronized (sync2_) {
			opens_ = -1;
		}
	}

	// loading event

	public void addLoadingListener(LoadingListener listener) {
		boolean completed;
		synchronized (sync_) {
			completed = completed_;
			loadingListeners_.add(LoadingListener.class, listener);
		}
		if (completed) {
			LoadingEvent e = new LoadingEvent(this, 1f, true);
			listener.loadingUpdated(e);
		}
	}

	public void removeLoadingListener(LoadingListener listener) {
		synchronized (sync_) {
			loadingListeners_.remove(LoadingListener.class, listener);
		}
	}

	protected void fireLoadingUpdated(float step, boolean completed) {
		LoadingListener[] listeners;
		synchronized (sync_) {
			if (completed) {
				completed_ = true;
			}
			listeners =
				loadingListeners_.getListeners(LoadingListener.class);
		}
		LoadingEvent e = new LoadingEvent(this, step, completed);
		for (LoadingListener listener : listeners) {
			listener.loadingUpdated(e);
		}
	}
}
